--[[
	Most instructions. Some are implemented in Core.lua to allow access to core variables without functions
	
	Mess with micro optimisation attempt and lot very similar function
]]

if not srb2gb then rawset(_G, "srb2gb", {}) end

local emu = srb2gb

--Put some global variables as local hoping for increased speed
local FLAGZMASK = emu.FLAGZMASK
local FLAGCMASK = emu.FLAGCMASK
local FLAGHMASK = emu.FLAGHMASK
local FLAGNMASK = emu.FLAGNMASK
local FLAGCLEAR = 0xFF00 -- Complement of the sum of all flags. When AND clear all flags(and remove bits that shouldn't be here too)

local REGA = emu.REGA
local REGC = emu.REGC
local REGAF = emu.REGAF
local REGAFBIS = REGAF + 1
local REGHL = emu.REGHL
local REGHLBIS = REGHL + 1
local REGBCBIS = emu.REGBC + 1
local REGDEBIS = emu.REGDE + 1

local REGFMASK = 0xFF
local REGHMASK = 0xFF00
local REGLMASK = 0xFF

local reg = emu.reg --Table abuse I guess

--Put some functions as local too(because global variables are too slow)
local getReg = emu.getReg
local getRegPair = emu.getRegPair
local readNext8 = emu.readNext8
local readNext16 = emu.readNext16
local read8 = emu.read8
local write8 = emu.write8
local setReg = emu.setReg
local setRegPair = emu.setRegPair
local nextCycle = emu.nextCycle
local addCycle = emu.addCycle
local setPC = emu.setPC
local addPC = emu.addPC
local call = emu.call
local push = emu.push
local pop = emu.pop

--Should be dropped or inlined as used in the most called instruction. And not portable
local function u8tos32(v)
	if v&0x80 then
		v = $ | 0xFFFFFF00
	end
	return v
end

local OPUNK = function() -- Throw instruction error here instead of testing at calling
	print("Unknown instruction "..emu.hexString(emu.current,2).." at "..emu.hexString(emu.pc - 1,4))
	emu.setState(emu.STATEERROR)
end

--Flow control
local opNOP = function() -- NOP
	return
end

--Arithmetics

--[[local oXORh = function() -- XOR (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = read8(reg[REGHLBIS])
	local f = 0
	a = a ^^ r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

local opORh = function() -- OR (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = read8(reg[REGHLBIS])
	local f = 0
	a = a | r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

local oANDh = function() -- AND (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = read8(reg[REGHLBIS])
	local f = FLAGHMASK
	a = a & r
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

local opCPL = function()
	reg[REGAFBIS] = ($ ^^ 0xFF00) | FLAGNMASK | FLAGHMASK --CPL can be done with XOR $FF
end

local oDECh = function() -- DEC (hl)
	local a = read8(emu.reg[REGHLBIS])
	local f = (reg[REGAFBIS] & FLAGCMASK) | FLAGNMASK
	a = ($ - 1)&0xFF
	write8(reg[REGHLBIS], a)
	if a == 0 then f = $ | FLAGZMASK
	elseif a&0xF == 0xF then f = $ | FLAGHMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

local oINCh = function() -- INC (hl)
	local a = read8(reg[REGHLBIS])
	local f = (reg[REGAFBIS] & FLAGCMASK)
	a = ($ + 1) & 0xFF
	write8(reg[REGHLBIS], a)
	if not (a&0xF) then f = $ | FLAGHMASK
		if a == 0 then f = $ | FLAGZMASK end
	end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end]]

local oANDn = function() -- AND n
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = readNext8()
	local f = FLAGHMASK
	a = a & b
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

local oORn = function() -- OR n
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = readNext8()
	local f = 0
	a = a | b
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

local oXORn = function() -- XOR n
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = readNext8()
	local f = 0
	a = a ^^ b
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

local opCPn = function() -- CP n
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = readNext8()
	local f = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then f = $ | FLAGHMASK end
	a = $ - b
	if not (a&0xFF) then f = $ | FLAGZMASK
	elseif a < 0 then f = $ | FLAGCMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

local opCPh = function() -- CP (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = read8(reg[REGHLBIS])
	local f = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then f = $ | FLAGHMASK end
	a = $ - b
	if not (a&0xFF) then f = $ | FLAGZMASK
	elseif a < 0 then f = $ | FLAGCMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

local oADDn = function() -- ADD A, n
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = readNext8()
	local f = 0
	if (a&0xF) + (b&0xF) >= 0x10 then f = FLAGHMASK end
	a = $ + b
	if a > 0xFF then f = $ | FLAGCMASK a = $ & 0xFF end
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oADDh = function() -- ADD A, (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = read8(reg[REGHLBIS])
	local f = 0
	if (a&0xF) + (b&0xF) >= 0x10 then f = FLAGHMASK end
	a = $ + b
	if a > 0xFF then f = $ | FLAGCMASK a = $ & 0xFF end
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oADCr = function(c) -- ADC A, r
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = getReg(c)
	local cy = false
	local f = 0
	if reg[REGAFBIS]&FLAGCMASK then
		if a & 0xF == 0xF then cy = true end
		a = $ + 1
	end
	if ((a&0xF) + (b&0xF) >= 0x10) or cy then f = FLAGHMASK end
	a = $ + b
	if (a > 0xFF) then f = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oADCn = function() -- ADC A, n
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = readNext8()
	local cy = false
	local f = 0
	if reg[REGAFBIS]&FLAGCMASK then
		if a & 0xF == 0xF then cy = true end
		a = $ + 1
	end
	if ((a&0xF) + (b&0xF) >= 0x10) or cy then f = FLAGHMASK end
	a = $ + b
	if (a > 0xFF) then f = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oADCh = function() -- ADC A, (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = read8(reg[REGHLBIS])
	local cy = false
	local f = 0
	if reg[REGAFBIS]&FLAGCMASK then
		if a & 0xF == 0xF then cy = true end
		a = $ + 1
	end
	if ((a&0xF) + (b&0xF) >= 0x10) or cy then f = FLAGHMASK end
	a = $ + b
	if (a > 0xFF) then f = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oSUBn = function() -- SUB A, n
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = readNext8()
	local f = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then f = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then f = $ | FLAGCMASK a = $ & 0xFF
	elseif not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oSUBh = function() -- SUB A, (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = read8(reg[REGHLBIS])
	local f = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then f = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then f = $ | FLAGCMASK a = $ & 0xFF
	elseif not a then f = $ | FLAGZMASK end
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oSBCr = function(c) -- SBC A, r
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = getReg(c)
	local cy = false
	local f = FLAGNMASK
	if reg[REGAFBIS]&FLAGCMASK then
		if a & 0xF == 0 then cy = true end
		a = $ - 1
	end
	if ((a&0xF) - (b&0xF) < 0) or cy then f = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then f = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oSBCn = function() -- SBC A, n
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = readNext8()
	local cy = false
	local f = FLAGNMASK
	if reg[REGAFBIS]&FLAGCMASK then
		if a & 0xF == 0 then cy = true end
		a = $ - 1
	end
	if ((a&0xF) - (b&0xF) < 0) or cy then f = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then f = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oSBCh = function() -- SBC A, (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local b = read8(reg[REGHLBIS])
	local cy = false
	local f = FLAGNMASK
	if reg[REGAFBIS]&FLAGCMASK then
		if a & 0xF == 0 then cy = true end
		a = $ - 1
	end
	if ((a&0xF) - (b&0xF) < 0) or cy then f = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then f = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

local oADDsp = function() -- ADD SP, n
	local a = emu.sp
	local b = u8tos32(readNext8())
	local f = 0
	if ((a&0xF) + (b&0xF)>0xF) then
		f = FLAGHMASK
	end
	if ((a&0xFF) + (b&0xFF)>0xFF) then
		f = $ | FLAGCMASK
	end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
	a = $ + b
	emu.sp = a & 0xFFFF
	nextCycle()
	nextCycle()
end

local opRRA = function()
	local a = (reg[REGAFBIS] >> 8)&0xFF
	if reg[REGAFBIS] & FLAGCMASK then
		a = $ | 0x100
	end
	reg[REGAFBIS] = ((a << 7)&0xFF00)
	if a&1 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
end

local opRLA = function()
	local a = (reg[REGAFBIS] >> 8)
	a = a << 1
	if reg[REGAFBIS] & FLAGCMASK then
		a = $ | 1
	end
	reg[REGAFBIS] = ((a << 8)&0xFF00)
	if a&0x100 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
end

local oRLCA = function()
	local a = (reg[REGAFBIS] >> 8)
	a = a << 1
	reg[REGAFBIS] = 0
	if a & 0x100 then
		a = $ | 1
		reg[REGAFBIS] = FLAGCMASK
	end
	reg[REGAFBIS] = ($ & REGFMASK) | ((a << 8)&0xFF00)
end

local oRRCA = function()
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local f = 0
	if a & 1 then
		a = $ | 0x100
		f = $ | FLAGCMASK
	end
	reg[REGAFBIS] = f | ((a << 7)&0xFF00)
end

--Transfer

local oLDrr = function(c) -- LD r, r
	setReg(c >> 3, getReg(c))
end

local oLDrn = function(c) -- LD r, n
	setReg(c>>3, readNext8())
end

local oLDhln = function() -- LD (hl), n
	write8(reg[REGHLBIS], readNext8())
end

local oLDrhl = function(c) -- LD r, (hl)
	setReg(c >> 3, read8(reg[REGHLBIS]))
end

local oLDhlr = function(c) -- LD (hl), r
	write8(reg[REGHLBIS], getReg(c))
end

local oLDIh = function() -- LDI (hl), a
	write8(reg[REGHLBIS], (reg[REGAFBIS] >> 8))
	reg[REGHLBIS] = ($ + 1)&0xFFFF
end

local oLDDh = function() -- LDD (hl), a
	write8(reg[REGHLBIS], (reg[REGAFBIS] >> 8))
	reg[REGHLBIS] = ($ - 1)&0xFFFF
end

local oLDIa = function() -- LDI a, (hl)
	reg[REGAFBIS] = ($ & REGFMASK) | (read8(reg[REGHLBIS]) << 8)
	reg[REGHLBIS] = ($ + 1)&0xFFFF
end

local oLDDa = function() -- LDD a, (hl)
	reg[REGAFBIS] = ($ & REGFMASK) | (read8(reg[REGHLBIS]) << 8)
	reg[REGHLBIS] = ($ - 1)&0xFFFF
end

local oLDHna = function() -- LD ($FF00 + n), a
	write8(0xFF00 + readNext8(), (reg[REGAFBIS] >> 8))
end

local oLDHca = function()  -- LD ($FF00 + c), a
	write8(0xFF00 | reg[REGBCBIS], (reg[REGAFBIS] >> 8))
end

local oLDHan = function() -- LD a, ($FF00 + n)
	reg[REGAFBIS] = ($ & REGFMASK) | (read8(0xFF00 + readNext8()) << 8)
end

local oLDHac = function() -- LD a, ($FF00 + c)
	reg[REGAFBIS] = ($ & REGFMASK) | (read8(0xFF00 | reg[REGBCBIS]) << 8)
end

local oLDnna = function() -- LD (nn), a
	write8(readNext16(), (reg[REGAFBIS] >> 8))
end

local oLDann = function() -- LD a, (nn)
	reg[REGAFBIS] = ($ & REGFMASK) | (read8(readNext16()) << 8)
end

local oLDsh = function() -- LD SP, HL
	emu.sp = reg[REGHLBIS]
	nextCycle()
end

local oLDns = function() -- LD nn, SP
	local a = emu.sp
	local b = readNext16()
	write8(b, a&0xFF)
	write8(b+1, a>>8)
end

local oLDhsn = function() -- LD HL, SP+e
	local a = emu.sp
	local b = u8tos32(readNext8())
	reg[REGAFBIS] = $ & FLAGCLEAR
	if ((a&0xF) + (b&0xF))>0xF then
		reg[REGAFBIS] = $ | FLAGHMASK
	end
	if ((a&0xFF) + (b&0xFF))>0xFF then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = $ + b
	reg[REGHLBIS] = a&0xFFFF
	nextCycle()
end

--CB

--CB insts

local cSWAP = function(current)
	local a = getReg(current)
	a = ($ >> 4) | (($ & 0x0F) << 4)
	setReg(current, a&0xFF)
	reg[REGAFBIS] = $ & FLAGCLEAR -- Reset every flags
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
end

local cSWAPh = function(current)
	local a = read8(reg[REGHLBIS])
	a = ($ >> 4) | (($ & 0x0F) << 4)
	write8(reg[REGHLBIS], a)
	reg[REGAFBIS] = $ & FLAGCLEAR -- Reset every flags
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
end

local cBIT0 = function(current) -- BIT 0, r ; Should perhaps move away optimized cb somewhere else
	reg[REGAFBIS] = ($ & (FLAGCLEAR | FLAGCMASK)) | FLAGHMASK --Clear everything but carry and set negative
	if not (getReg(current) & 1) then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
end

local cBIT7 = function(current) -- BIT 7, r
	reg[REGAFBIS] = ($ & (FLAGCLEAR | FLAGCMASK)) | FLAGHMASK --Clear everything but carry and set negative
	if not (getReg(current) & 0x80) then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
end

local cbBIT = function(current)
	reg[REGAFBIS] = ($ & (FLAGCLEAR | FLAGCMASK)) | FLAGHMASK --Clear everything but carry and set negative
	emu.setFlag(FLAGZMASK, not (getReg(current) & (1<<((current>>3)&0x7))) )
end

local cBITh = function(current) --BIT b, (hl)
	reg[REGAFBIS] = ($ & (FLAGCLEAR | FLAGCMASK)) | FLAGHMASK --Clear everything but carry and set negative
	emu.setFlag(FLAGZMASK, not (read8(reg[REGHLBIS]) & (1<<((current>>3)&0x7))) )
end

local cbSET = function(current) -- SET b, r
	setReg(current, getReg(current) | (1<<((current>>3)&0x7)) ) --Nice instruction composed of 1 line
end

local cSET0 = function(current) -- SET 0, r
	setReg(current, getReg(current) | 1) --Nice instruction composed of 1 line
end
local cSET7 = function(current) -- SET 7, r
	setReg(current, getReg(current) | 0x80) --Nice instruction composed of 1 line
end

local cSETh = function(current) --SET b, (hl)
	write8(reg[REGHLBIS], read8(reg[REGHLBIS]) | (1<<((current>>3)&0x7)) ) --Nice instruction composed of 1 line
end

local cbRES = function(current) --RES b, r
	setReg(current, getReg(current) & (~(1<<((current>>3)&0x7))) )
end

local cRES0 = function(current) --RES 0, r
	setReg(current, getReg(current) & 0xFE)
end

local cRES7 = function(current) --RES 7, r
	setReg(current, getReg(current) & 0x7F)
end

local cRESh = function(current) --RES b, (hl)
	write8(reg[REGHLBIS], read8(reg[REGHLBIS]) & (~(1<<((current>>3)&0x7))) )
end

local cbSLA = function(current)
	local a = getReg(current)
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x80 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = ($ << 1) & 0xFF
	if not a then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
	setReg(current, a)
end

local cbSRA = function(current)
	local a = getReg(current)
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x1 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = ($ >> 1) | ($ & 0x80)
	if not a then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
	setReg(current, a)
end

local cSLAh = function(current)
	local a = read8(reg[REGHLBIS])
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x80 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = ($ << 1) & 0xFF
	if not a then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
	write8(reg[REGHLBIS], a)
end

local cSRAh = function(current)
	local a = read8(reg[REGHLBIS])
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x1 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = ($ >> 1) | ($ & 0x80)
	if not a then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
	write8(reg[REGHLBIS], a)
end

local cbSRL = function(current)
	local a = getReg(current)
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x1 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = a >> 1
	if not a then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
	setReg(current, a)
end

local cSRLh = function(current)
	local a = read8(reg[REGHLBIS])
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x1 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = a >> 1
	if not a then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
	write8(reg[REGHLBIS], a)
end

local cbRL = function(current)
	local a = getReg(current)
	a = $ << 1
	if reg[REGAFBIS] & FLAGCMASK then
		a = $ | 1
	end
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x100 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = $ & 0xFF
	if not a then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
	emu.setReg(current, a)
end

local cbRR = function(current)
	local a = getReg(current)
	if reg[REGAFBIS] & FLAGCMASK then
		a = $ | 0x100
	end
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x1 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = $ >> 1
	if not a then
		emu.reg[REGAFBIS] = $ | FLAGZMASK
	end
	emu.setReg(current, a)
end

local cbRLh = function(current)
	local a = read8(reg[REGHLBIS])
	a = $ << 1
	if reg[REGAFBIS] & FLAGCMASK then
		a = $ | 1
	end
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x100 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = $ & 0xFF
	if not a then
		reg[REGAFBIS] = $ | FLAGZMASK
	end
	write8(reg[REGHLBIS], a)
end

local cbRRh = function(current)
	local a = read8(reg[REGHLBIS])
	if reg[REGAFBIS] & FLAGCMASK then
		a = $ | 0x100
	end
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a&0x1 then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = $ >> 1
	if not a then
		emu.reg[REGAFBIS] = $ | FLAGZMASK
	end
	write8(reg[REGHLBIS], a)
end

local cbRLC = function(current)
	local a = emu.getReg(current)
	a = a << 1
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a & 0x100 then
		a = $ | 1
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = a & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	emu.setReg(current, a)
end

local cbRRC = function(current)
	local a = emu.getReg(current)
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a & 1 then
		a = $ | 0x100
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = (a >> 1) & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	setReg(current, a)
end

local cRLCh = function(current)
	local a = read8(reg[REGHLBIS])
	a = a << 1
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a & 0x100 then
		a = $ | 1
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = a & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	write8(reg[REGHLBIS], a)
end

local cRRCh = function(current)
	local a = read8(reg[REGHLBIS])
	reg[REGAFBIS] = $ & FLAGCLEAR
	if a & 1 then
		a = $ | 0x100
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	a = (a >> 1) & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	write8(reg[REGHLBIS], a)
end

local cb = {
	       cbRLC, cbRLC, cbRLC, cbRLC, cbRLC, cRLCh, cbRLC, cbRRC, cbRRC, cbRRC, cbRRC, cbRRC, cbRRC, cRRCh, cbRRC, --0x
	cbRL,  cbRL,  cbRL,  cbRL,  cbRL,  cbRL,  cbRLh, cbRL,  cbRR,  cbRR,  cbRR,  cbRR,  cbRR,  cbRR,  cbRRh, cbRR,  --1x
	cbSLA, cbSLA, cbSLA, cbSLA, cbSLA, cbSLA, cSLAh, cbSLA, cbSRA, cbSRA, cbSRA, cbSRA, cbSRA, cbSRA, cSRAh, cbSRA, --2x
	cSWAP, cSWAP, cSWAP, cSWAP, cSWAP, cSWAP, cSWAPh,cSWAP, cbSRL, cbSRL, cbSRL, cbSRL, cbSRL, cbSRL, cSRLh, cbSRL, --3x
	cBIT0, cBIT0, cBIT0, cBIT0, cBIT0, cBIT0, cBITh, cBIT0, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cBITh, cbBIT, --4x
	cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cBITh, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cBITh, cbBIT, --5x
	cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cBITh, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cBITh, cbBIT, --6x
	cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cbBIT, cBITh, cbBIT, cBIT7, cBIT7, cBIT7, cBIT7, cBIT7, cBIT7, cBITh, cBIT7, --7x
	cRES0, cRES0, cRES0, cRES0, cRES0, cRES0, cRESh, cRES0, cbRES, cbRES, cbRES, cbRES, cbRES, cbRES, cRESh, cbRES, --8x
	cbRES, cbRES, cbRES, cbRES, cbRES, cbRES, cRESh, cbRES, cbRES, cbRES, cbRES, cbRES, cbRES, cbRES, cRESh, cbRES, --9x
	cbRES, cbRES, cbRES, cbRES, cbRES, cbRES, cRESh, cbRES, cbRES, cbRES, cbRES, cbRES, cbRES, cbRES, cRESh, cbRES, --Ax
	cbRES, cbRES, cbRES, cbRES, cbRES, cbRES, cRESh, cbRES, cRES7, cRES7, cRES7, cRES7, cRES7, cRES7, cRESh, cRES7, --Bx
	cSET0, cSET0, cSET0, cSET0, cSET0, cSET0, cSETh, cSET0, cbSET, cbSET, cbSET, cbSET, cbSET, cbSET, cSETh, cbSET, --Cx
	cbSET, cbSET, cbSET, cbSET, cbSET, cbSET, cSETh, cbSET, cbSET, cbSET, cbSET, cbSET, cbSET, cbSET, cSETh, cbSET, --Dx
	cbSET, cbSET, cbSET, cbSET, cbSET, cbSET, cSETh, cbSET, cbSET, cbSET, cbSET, cbSET, cbSET, cbSET, cSETh, cbSET, --Ex
	cbSET, cbSET, cbSET, cbSET, cbSET, cbSET, cSETh, cbSET, cSET7, cSET7, cSET7, cSET7, cSET7, cSET7, cSETh, cSET7, --Fx
}
cb[0] = cbRLC

local cbUNK = function(current)
	print("Unknown CB inst: "..emu.hexString(current,2).." at "..emu.hexString(emu.pc - 2, 4))
	emu.setState(emu.STATEERROR)
end

local opCB = function() -- CBxx
	local op = readNext8()
	cb[op](op)
end

--Put it in the op array. False value are defined later to reduce number of function
local op = {
	       false, false, false, false, false, false, oRLCA, oLDns, false, false, false, false, false, false, oRRCA, --0x
	false, false, false, false, false, false, false, opRLA, false, false, false, false, false, false, false, opRRA, --1x
	false, false, oLDIh, false, false, false, false, false, false, false, oLDIa, false, false, false, false, false, --2x
	false, false, oLDDh, false, false, false, oLDhln,false, false, false, oLDDa, false, false, false, false, false, --3x
	opNOP, oLDrr, oLDrr, oLDrr, oLDrr, oLDrr, oLDrhl,false, oLDrr, opNOP, oLDrr, oLDrr, oLDrr, oLDrr, oLDrhl,false, --4x
	oLDrr, oLDrr, opNOP, oLDrr, oLDrr, oLDrr, oLDrhl,false, oLDrr, oLDrr, oLDrr, opNOP, oLDrr, oLDrr, oLDrhl,false, --5x
	oLDrr, oLDrr, oLDrr, oLDrr, opNOP, oLDrr, oLDrhl,false, oLDrr, oLDrr, oLDrr, oLDrr, oLDrr, opNOP, oLDrhl,false, --6x
	oLDhlr,oLDhlr,oLDhlr,oLDhlr,oLDhlr,oLDhlr,false, false, false, false, false, false, false, false, false, opNOP, --7x
	false, false, false, false, false, false, oADDh, false, oADCr, oADCr, oADCr, oADCr, oADCr, oADCr, oADCh, oADCr, --8x
	false, false, false, false, false, false, oSUBh, false, oSBCr, oSBCr, oSBCr, oSBCr, oSBCr, oSBCr, oSBCh, oSBCr, --9x
	false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, --Ax
	false, false, false, false, false, false, false, false, false, false, false, false, false, false, opCPh, false, --Bx
	false, false, false, false, false, false, oADDn, false, false, false, false, opCB,  false, false, oADCn, false, --Cx
	false, false, false, OPUNK, false, false, oSUBn, false, false, false, false, OPUNK, false, OPUNK, oSBCn, false, --Dx
	oLDHna,false, oLDHca,OPUNK, OPUNK, false, oANDn, false, oADDsp,false, oLDnna,OPUNK, OPUNK, OPUNK, oXORn, false, --Ex
	oLDHan,false, oLDHac,false, OPUNK, false, oORn,  false, oLDhsn,oLDsh, oLDann,false, OPUNK, OPUNK, opCPn, false  --Fx
}

op[0] = opNOP
--Thanks Lua and it's limitation in number of local variables

-- Various LD

op[0x01] = function() -- LD BC, nn
	reg[REGBCBIS] = readNext16()
end

op[0x11] = function() -- LD DE, nn
	reg[REGDEBIS] = readNext16()
end

op[0x21] = function() -- LD hl, dd
	reg[REGHLBIS] = readNext16()
end

op[0x31] = function() -- LD sp, nn
	emu.sp = readNext16()
end

op[0x0A] = function() -- LD a, (BC)
	reg[REGAFBIS] = ($ & REGFMASK) | (read8(reg[REGBCBIS]) << 8)
end

op[0x1A] = function() -- LD a, (DE)
	reg[REGAFBIS] = ($ & REGFMASK) | (read8(reg[REGDEBIS]) << 8)
end

op[0x02] = function() -- LD (BC), a
	write8(reg[REGBCBIS], reg[REGAFBIS] >> 8)
end

op[0x12] = function() -- LD (DE), a
	write8(reg[REGDEBIS], reg[REGAFBIS] >> 8)
end

op[0x7E] = function() -- LD a, (hl)
	reg[REGAFBIS] = ($ & REGFMASK) | (read8(reg[REGHLBIS]) << 8)
end

op[0x77] = function() -- LD (hl), a
	write8(reg[REGHLBIS], reg[REGAFBIS] >> 8)
end

-- INC
op[0x04] = function() -- INC b
	local a = reg[REGBCBIS] >> 8
	a = ($ + 1)&0xFF
	reg[REGBCBIS] = $ & REGLMASK | (a<<8)
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) )
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	if not (a&0xF) then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x0C] = function() -- INC c
	local a = reg[REGBCBIS] & 0xFF
	a = ($ + 1)&0xFF
	reg[REGBCBIS] = $ & REGHMASK | a
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) )
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	if not (a&0xF) then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x14] = function() -- INC d
	local a = reg[REGDEBIS] >> 8
	a = ($ + 1)&0xFF
	reg[REGDEBIS] = $ & REGLMASK | (a<<8)
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) )
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	if not (a&0xF) then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x1C] = function() -- INC e
	local a = reg[REGDEBIS] & 0xFF
	a = ($ + 1)&0xFF
	reg[REGDEBIS] = $ & REGHMASK | a
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) )
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	if not (a&0xF) then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x24] = function() -- INC h
	local a = reg[REGHLBIS] >> 8
	a = ($ + 1)&0xFF
	reg[REGHLBIS] = $ & REGLMASK | (a<<8)
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) )
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	if not (a&0xF) then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x2C] = function() -- INC l
	local a = reg[REGHLBIS] & 0xFF
	a = ($ + 1)&0xFF
	reg[REGHLBIS] = $ & REGHMASK | a
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) )
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	if not (a&0xF) then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x34] = function() -- INC (hl)
	local a = read8(reg[REGHLBIS])
	local f = (reg[REGAFBIS] & FLAGCMASK)
	a = ($ + 1) & 0xFF
	write8(reg[REGHLBIS], a)
	if not (a&0xF) then f = $ | FLAGHMASK
		if a == 0 then f = $ | FLAGZMASK end
	end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

op[0x3C] = function() -- INC a
	local a = reg[REGAFBIS] >> 8
	a = ($ + 1)&0xFF
	reg[REGAFBIS] = ($ & FLAGCMASK) | (a<<8)
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	if not (a&0xF) then reg[REGAFBIS] = $ | FLAGHMASK end
end

-- DEC
op[0x05] = function() -- DEC b
	local a = reg[REGBCBIS] >> 8
	a = ($ - 1) &0xFF
	reg[REGBCBIS] = ($ & REGLMASK) | (a<<8)
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) ) | FLAGNMASK
	if a == 0 then reg[REGAFBIS] = $ | FLAGZMASK end
	if a&0xF == 0xF then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x0D] = function() -- DEC c
	local a = reg[REGBCBIS] & 0xFF
	a = ($ - 1)&0xFF
	reg[REGBCBIS] = ($ & REGHMASK) | a
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) ) | FLAGNMASK
	if a == 0 then reg[REGAFBIS] = $ | FLAGZMASK end
	if a&0xF == 0xF then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x15] = function() -- DEC d
	local a = reg[REGDEBIS] >> 8
	a = ($ - 1)&0xFF
	reg[REGDEBIS] = ($ & REGLMASK) | (a<<8)
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) ) | FLAGNMASK
	if a == 0 then reg[REGAFBIS] = $ | FLAGZMASK end
	if a&0xF == 0xF then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x1D] = function() -- DEC e
	local a = reg[REGDEBIS] & 0xFF
	a = ($ - 1)&0xFF
	reg[REGDEBIS] = ($ & REGHMASK) | a
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) ) | FLAGNMASK
	if a == 0 then reg[REGAFBIS] = $ | FLAGZMASK end
	if a&0xF == 0xF then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x25] = function() -- DEC h
	local a = reg[REGHLBIS] >> 8
	a = ($ - 1)&0xFF
	reg[REGHLBIS] = ($ & REGLMASK) | (a<<8)
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) ) | FLAGNMASK
	if a == 0 then reg[REGAFBIS] = $ | FLAGZMASK end
	if a&0xF == 0xF then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x2D] = function() -- DEC l
	local a = reg[REGHLBIS] & 0xFF
	a = ($ - 1)&0xFF
	reg[REGHLBIS] = ($ & REGHMASK) | a
	reg[REGAFBIS] = ($ & (FLAGCLEAR|FLAGCMASK) ) | FLAGNMASK
	if a == 0 then reg[REGAFBIS] = $ | FLAGZMASK end
	if a&0xF == 0xF then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x35] = function() -- DEC (hl)
	local a = read8(emu.reg[REGHLBIS])
	local f = (reg[REGAFBIS] & FLAGCMASK) | FLAGNMASK
	a = ($ - 1)&0xFF
	write8(reg[REGHLBIS], a)
	if a == 0 then f = $ | FLAGZMASK
	elseif a&0xF == 0xF then f = $ | FLAGHMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

op[0x3D] = function() -- DEC a
	local a = reg[REGAFBIS] >> 8
	a = ($ - 1)&0xFF
	reg[REGAFBIS] = ($ & FLAGCMASK) | (a<<8) | FLAGNMASK
	if a == 0 then reg[REGAFBIS] = $ | FLAGZMASK end
	if a&0xF == 0xF then reg[REGAFBIS] = $ | FLAGHMASK end
end

op[0x0B] = function() -- DEC BC
	reg[REGBCBIS] = ($ - 1)&0xFFFF
	nextCycle()
end

op[0x1B] = function() -- DEC DE
	reg[REGDEBIS] = ($ - 1)&0xFFFF
	nextCycle()
end

op[0x2B] = function() -- DEC HL
	reg[REGHLBIS] = ($ - 1)&0xFFFF
	nextCycle()
end

op[0x03] = function() -- INC BC
	reg[REGBCBIS] = ($ + 1)&0xFFFF
	nextCycle()
end

op[0x13] = function() -- INC DE
	reg[REGDEBIS] = ($ + 1)&0xFFFF
	nextCycle()
end

op[0x23] = function() -- INC HL
	reg[REGHLBIS] = ($ + 1)&0xFFFF
	nextCycle()
end

op[0x2F] = function() -- CPL
	reg[REGAFBIS] = ($ ^^ 0xFF00) | FLAGNMASK | FLAGHMASK --CPL can be done with XOR $FF
end

-- ADD
op[0x09] = function() -- ADD hl, bc
	local a = reg[REGHLBIS]
	local b = reg[REGBCBIS]
	
	reg[REGAFBIS] = $ & (FLAGCLEAR|FLAGZMASK)
	if (a&0xFFF) + (b&0xFFF) >= 0x1000 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	
	reg[REGHLBIS] = a & 0xFFFF
	if a > 0xFFFF then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	nextCycle()
end

op[0x19] = function() -- ADD hl, de
	local a = reg[REGHLBIS]
	local b = reg[REGDEBIS]
	
	reg[REGAFBIS] = $ & (FLAGCLEAR|FLAGZMASK)
	if (a&0xFFF) + (b&0xFFF) >= 0x1000 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	
	reg[REGHLBIS] = a & 0xFFFF
	if a > 0xFFFF then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	nextCycle()
end

op[0x29] = function() -- ADD hl, hl
	local a = reg[REGHLBIS]
	
	reg[REGAFBIS] = $ & (FLAGCLEAR|FLAGZMASK)
	if a&0x800 then reg[REGAFBIS] = $ | FLAGHMASK end --Carry on 12th bit if 11th is set should work
	if a & 0x8000 then reg[REGAFBIS] = $ | FLAGCMASK end
	
	reg[REGHLBIS] = (a + a)&0xFFFF
	nextCycle()
end

op[0x39] = function() -- ADD hl, sp
	local a = reg[REGHLBIS]
	local b = emu.sp
	
	reg[REGAFBIS] = $ & (FLAGCLEAR|FLAGZMASK)
	if (a&0xFFF) + (b&0xFFF) >= 0x1000 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	
	reg[REGHLBIS] = a & 0xFFFF
	if a > 0xFFFF then
		reg[REGAFBIS] = $ | FLAGCMASK
	end
	nextCycle()
end

op[0x80] = function() -- ADD A, B
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGBCBIS] >> 8
	reg[REGAFBIS] = 0 -- Why keep A if it's to overwrite it after
	if (a&0xF) + (b&0xF) >= 0x10 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	if a > 0xFF then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x82] = function() -- ADD A, D
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGDEBIS] >> 8
	reg[REGAFBIS] = 0
	if (a&0xF) + (b&0xF) >= 0x10 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	if a > 0xFF then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x84] = function() -- ADD A, H
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGHLBIS] >> 8
	reg[REGAFBIS] = 0
	if (a&0xF) + (b&0xF) >= 0x10 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	if a > 0xFF then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x81] = function() -- ADD A, C
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGBCBIS]&0xFF
	reg[REGAFBIS] = 0
	if (a&0xF) + (b&0xF) >= 0x10 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	if a > 0xFF then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x83] = function() -- ADD A, E
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGDEBIS]&0xFF
	reg[REGAFBIS] = 0
	if (a&0xF) + (b&0xF) >= 0x10 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	if a > 0xFF then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x85] = function() -- ADD A, L
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGHLBIS]&0xFF
	reg[REGAFBIS] = 0
	if (a&0xF) + (b&0xF) >= 0x10 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ + b
	if a > 0xFF then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x87] = function() -- ADD a, a ; Or a*2
	local a = reg[REGAFBIS] >> 8
	reg[REGAFBIS] = ((a << 9)&0xFF00)
	if a&8 then reg[REGAFBIS] = $ | FLAGHMASK end --Same as for ADD hl, hl
	if a & 0x80 then reg[REGAFBIS] = $ | FLAGCMASK end
	if not (a&0x7F) then reg[REGAFBIS] = $ | FLAGZMASK end
end

op[0x90] = function() -- SUB A, B
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGBCBIS] >> 8
	reg[REGAFBIS] = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x91] = function() -- SUB A, C
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGBCBIS]&0xFF
	reg[REGAFBIS] = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x92] = function() -- SUB A, D
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGDEBIS] >> 8
	reg[REGAFBIS] = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x93] = function() -- SUB A, E
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGDEBIS]&0xFF
	reg[REGAFBIS] = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x94] = function() -- SUB A, H
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGHLBIS] >> 8
	reg[REGAFBIS] = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x95] = function() -- SUB A, L
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGHLBIS]&0xFF
	reg[REGAFBIS] = FLAGNMASK
	if (a&0xF) - (b&0xF) < 0 then reg[REGAFBIS] = $ | FLAGHMASK end
	a = $ - b
	if a < 0 then reg[REGAFBIS] = $ | FLAGCMASK end
	a = $ & 0xFF
	if not a then reg[REGAFBIS] = $ | FLAGZMASK end
	reg[REGAFBIS] = ($ & REGFMASK) | (a << 8)
end

op[0x97] = function() --SUB a,a ; Redundant with "XOR a" so usually not used
	reg[REGAFBIS] = FLAGNMASK | FLAGZMASK
end

op[0xA0] = function(c) -- AND b
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGBCBIS] >> 8
	local f = FLAGHMASK
	a = a & r
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xA1] = function(c) -- AND c
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGBCBIS]&0xFF
	local f = FLAGHMASK
	a = a & r
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xA2] = function(c) -- AND d
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGDEBIS] >> 8
	local f = FLAGHMASK
	a = a & r
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xA3] = function(c) -- AND e
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGDEBIS]&0xFF
	local f = FLAGHMASK
	a = a & r
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xA4] = function(c) -- AND h
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGHLBIS] >> 8
	local f = FLAGHMASK
	a = a & r
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xA5] = function(c) -- AND l
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGHLBIS]&0xFF
	local f = FLAGHMASK
	a = a & r
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xA6] = function() -- AND (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = read8(reg[REGHLBIS])
	local f = FLAGHMASK
	a = a & r
	if not a then
		f = $ | FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xA7] = function() -- AND a: Aka test A
	if reg[REGAFBIS] >> 8 then
		reg[REGAFBIS] = ($ & FLAGCLEAR) | FLAGHMASK
	else
		reg[REGAFBIS] = ($ & FLAGCLEAR) | FLAGHMASK | FLAGZMASK
	end
end

op[0xA8] = function(c) -- XOR b
	local a = reg[REGAFBIS] >> 8
	local r = reg[REGBCBIS] >> 8
	local f = 0
	a = a ^^ r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xA9] = function(c) -- XOR c
	local a = reg[REGAFBIS] >> 8
	local r = reg[REGBCBIS]&0xFF
	local f = 0
	a = a ^^ r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xAA] = function(c) -- XOR d
	local a = reg[REGAFBIS] >> 8
	local r = reg[REGDEBIS] >> 8
	local f = 0
	a = a ^^ r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xAB] = function(c) -- XOR e
	local a = reg[REGAFBIS] >> 8
	local r = reg[REGDEBIS]&0xFF
	local f = 0
	a = a ^^ r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xAC] = function(c) -- XOR h
	local a = reg[REGAFBIS] >> 8
	local r = reg[REGHLBIS] >> 8
	local f = 0
	a = a ^^ r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xAD] = function(c) -- XOR l
	local a = reg[REGAFBIS] >> 8
	local r = reg[REGHLBIS]&0xFF
	local f = 0
	a = a ^^ r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xAE] = function() -- XOR (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = read8(reg[REGHLBIS])
	local f = 0
	a = a ^^ r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xAF] = function() --XOR a ; Aka a = 0
	reg[REGAFBIS] = FLAGZMASK
end

op[0xB0] = function(c) -- OR b
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGBCBIS] >> 8
	local f = 0
	a = a | r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xB1] = function(c) -- OR c
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGBCBIS]&0xFF
	local f = 0
	a = a | r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xB2] = function(c) -- OR d
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGDEBIS] >> 8
	local f = 0
	a = a | r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xB3] = function(c) -- OR e
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGDEBIS]&0xFF
	local f = 0
	a = a | r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xB4] = function(c) -- OR h
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGHLBIS] >> 8
	local f = 0
	a = a | r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xB5] = function(c) -- OR l
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = reg[REGHLBIS]&0xFF
	local f = 0
	a = a | r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xB6] = function() -- OR (hl)
	local a = (reg[REGAFBIS] >> 8)&0xFF
	local r = read8(reg[REGHLBIS])
	local f = 0
	a = a | r
	if not a then
		f = FLAGZMASK
	end
	reg[REGAFBIS] = (a << 8) | f
end

op[0xB7] = function() -- OR a ; Equivalent to "AND a" but used less often
	if reg[REGAFBIS] >> 8 then
		reg[REGAFBIS] = $ & FLAGCLEAR
	else
		reg[REGAFBIS] = ($ & FLAGCLEAR) | FLAGZMASK
	end
end

op[0xB8] = function(c) -- CP b
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGBCBIS] >> 8
	local f = FLAGNMASK
	if (a&0xF) < (b&0xF) then f = $ | FLAGHMASK end
	if a == b then f = $ | FLAGZMASK
	elseif a < b then f = $ | FLAGCMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

op[0xB9] = function(c) -- CP c
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGBCBIS]&0xFF
	local f = FLAGNMASK
	if (a&0xF) < (b&0xF) then f = $ | FLAGHMASK end
	if a == b then f = $ | FLAGZMASK
	elseif a < b then f = $ | FLAGCMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

op[0xBA] = function(c) -- CP d
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGDEBIS] >> 8
	local f = FLAGNMASK
	if (a&0xF) < (b&0xF) then f = $ | FLAGHMASK end
	if a == b then f = $ | FLAGZMASK
	elseif a < b then f = $ | FLAGCMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

op[0xBB] = function(c) -- CP e
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGDEBIS]&0xFF
	local f = FLAGNMASK
	if (a&0xF) < (b&0xF) then f = $ | FLAGHMASK end
	if a == b then f = $ | FLAGZMASK
	elseif a < b then f = $ | FLAGCMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

op[0xBC] = function(c) -- CP h
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGHLBIS] >> 8
	local f = FLAGNMASK
	if (a&0xF) < (b&0xF) then f = $ | FLAGHMASK end
	if a == b then f = $ | FLAGZMASK
	elseif a < b then f = $ | FLAGCMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

op[0xBD] = function(c) -- CP l
	local a = reg[REGAFBIS] >> 8
	local b = reg[REGHLBIS]&0xFF
	local f = FLAGNMASK
	if (a&0xF) < (b&0xF) then f = $ | FLAGHMASK end
	if a == b then f = $ | FLAGZMASK
	elseif a < b then f = $ | FLAGCMASK end
	reg[REGAFBIS] = ($ & FLAGCLEAR) | f
end

op[0xBF] = function() --CP a ; An useless OPcode again I guess. Compare but result is constant
	reg[REGAFBIS] = ($ & FLAGCLEAR) | FLAGNMASK | FLAGZMASK
end

--RST
op[0xC7] = function() -- RST 0
	call(0)
	nextCycle()
end

op[0xCF] = function() -- RST 8
	call(0x08)
	nextCycle()
end

op[0xD7] = function() -- RST 10
	call(0x10)
	nextCycle()
end

op[0xDF] = function() -- RST 18
	call(0x18)
	nextCycle()
end

op[0xE7] = function() -- RST 20
	call(0x20)
	nextCycle()
end

op[0xEF] = function() -- RST 28
	call(0x28)
	nextCycle()
end

op[0xF7] = function() -- RST 30
	call(0x30)
	nextCycle()
end

op[0xFF] = function() -- RST 38
	if emu.debug.value then
		print("Warning: RST 38")
		emu.setState(emu.STATEERROR)
	end
	call(0x38)
	nextCycle()
end

op[0x33] = function() -- INC sp
	emu.sp = ($ + 1)&0xFFFF
	nextCycle()
end

op[0x3B] = function() -- DEC sp
	emu.sp = ($ - 1)&0xFFFF
	nextCycle()
end

-- POP
op[0xC1] = function() -- POP bc
	reg[REGBCBIS] = pop()
end

op[0xD1] = function() -- POP de
	reg[REGDEBIS] = pop()
end

op[0xE1] = function() -- POP hl
	reg[REGHLBIS] = pop()
end

op[0xF1] = function() -- POP af ; Need to adapt as it's AF
	reg[REGAFBIS] = pop()&0xFFF0
end

--PUSH
op[0xC5] = function() -- PUSH bc
	push(reg[REGBCBIS])
	nextCycle()
end

op[0xD5] = function() -- PUSH de
	push(reg[REGDEBIS])
	nextCycle()
end

op[0xE5] = function() -- PUSH hl
	push(reg[REGHLBIS])
	nextCycle()
end

op[0xF5] = function() -- PUSH af
	push(reg[REGAFBIS])
	nextCycle()
end

-- LD r, r

op[0x47] = function() -- LD b, a
	reg[REGBCBIS] = ($ & REGFMASK) | (reg[REGAFBIS]&0xFF00)
end

op[0x4F] = function() -- LD c, a
	reg[REGBCBIS] = ($ & REGHMASK) | (reg[REGAFBIS]>>8)
end

op[0x57] = function() -- LD d, a
	reg[REGDEBIS] = ($ & REGFMASK) | (reg[REGAFBIS]&0xFF00)
end

op[0x5F] = function() -- LD e, a
	reg[REGDEBIS] = ($ & REGHMASK) | (reg[REGAFBIS]>>8)
end

op[0x67] = function() -- LD h, a
	reg[REGHLBIS] = ($ & REGFMASK) | (reg[REGAFBIS]&0xFF00)
end

op[0x6F] = function() -- LD l, a
	reg[REGHLBIS] = ($ & REGHMASK) | (reg[REGAFBIS]>>8)
end

op[0x78] = function() -- LD a, b
	reg[REGAFBIS] = ($ & REGFMASK) | (reg[REGBCBIS]&0xFF00)
end

op[0x79] = function() -- LD a, c
	reg[REGAFBIS] = ($ & REGFMASK) | ((reg[REGBCBIS]&0xFF) << 8)
end

op[0x7A] = function() -- LD a, d
	reg[REGAFBIS] = ($ & REGFMASK) | (reg[REGDEBIS]&0xFF00)
end

op[0x7B] = function() -- LD a, e
	reg[REGAFBIS] = ($ & REGFMASK) | ((reg[REGDEBIS]&0xFF) << 8)
end

op[0x7C] = function() -- LD a, h
	reg[REGAFBIS] = ($ & REGFMASK) | (reg[REGHLBIS]&0xFF00)
end

op[0x7D] = function() -- LD a, l
	reg[REGAFBIS] = ($ & REGFMASK) | ((reg[REGHLBIS]&0xFF) << 8)
end

-- LD r, n

op[0x06] = function(c) -- LD b, n
	reg[REGBCBIS] = ($ & REGFMASK) | (readNext8() << 8)
end

op[0x16] = function(c) -- LD d, n
	reg[REGDEBIS] = ($ & REGFMASK) | (readNext8() << 8)
end

op[0x26] = function(c) -- LD h, n
	reg[REGHLBIS] = ($ & REGFMASK) | (readNext8() << 8)
end

op[0x3E] = function(c) -- LD h, n
	reg[REGAFBIS] = ($ & REGFMASK) | (readNext8() << 8)
end

op[0x0E] = function(c) -- LD c, n
	reg[REGBCBIS] = ($ & REGHMASK) | (readNext8())
end

op[0x1E] = function(c) -- LD e, n
	reg[REGDEBIS] = ($ & REGHMASK) | (readNext8())
end

op[0x2E] = function(c) -- LD c, n
	reg[REGHLBIS] = ($ & REGHMASK) | (readNext8())
end

op[0x37] = function() -- SCF
	reg[REGAFBIS] = ($ & (~(FLAGNMASK|FLAGHMASK))) | FLAGCMASK
end

op[0x3F] = function() -- SCF
	reg[REGAFBIS] = ($ & (~(FLAGNMASK|FLAGHMASK))) ^^ FLAGCMASK
end

op[0x27] = function() -- DAA ; The worse instruction probably. Thanks open source emu for easier way to do it
	local f = reg[REGAFBIS]
	local a = (f >> 8)
	if f & FLAGNMASK then
		if f & FLAGHMASK then a = (a - 0x06) & 0xFF end
		if f & FLAGCMASK then a = $ - 0x60 end
	else
		if (f & FLAGHMASK) or ((a & 0xF) > 9) then a = $ + 6 end
		if (f & FLAGCMASK) or (a > 0x9F) then a = $ + 0x60 end
	end
	f = $ & (FLAGNMASK|FLAGCMASK)
	if a & 0x100 then f = $ | FLAGCMASK a = $ & 0xFF end
	if not a then f = $ | FLAGZMASK end
	reg[REGAFBIS] = f | (a << 8)
end

emu.op = op

-- Add JP, JR, EI, DI, HALT, STOP...
emu.initCoreInst()